#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""
import argparse
from collections.abc import Iterable, Mapping
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Any, Final, Optional

from pdm_pfsc.abstractions import ConfigHolder
from pdm_pfsc.config import (
    ConfigAccessor,
    ConfigNamespace,
    ConfigSection,
    IsMissing,
)
from pdm_pfsc.logging import traced_function


class _StringEnum(str, Enum):
    """"""

    # Justification: Zen of python: Explicit is better than implicit
    pass  # pylint: disable=W0107


class _ConfigKeys(_StringEnum):
    """"""

    VERSION: Final[str] = "version"
    VERSION_SOURCE: Final[str] = "source"
    VERSION_SOURCE_FILE_PATH: Final[str] = "path"
    BUILD_BACKEND: Final[str] = "build-backend"
    VCS_PROVIDER: Final[str] = "provider"
    PROJECT_METADATA: Final[str] = "project"


VERSION_CONFIG_KEY_NAME: Final[str] = _ConfigKeys.VERSION


class _ConfigValues(_StringEnum):
    """"""

    VERSION_SOURCE_FILE: Final[str] = "file"
    VERSION_SOURCE_SCM: Final[str] = "scm"
    DEPRECATED_BUILD_BACKEND_PDM_PEP517_API: Final[str] = "pdm.pep517.api"
    BUILD_BACKEND_PDM_BACKEND: Final[str] = "pdm.backend"


class _ConfigSections(_StringEnum):
    """"""

    PDM_BUMP: Final[str] = "pdm_bump"
    PDM_BUMP_VCS: Final[str] = "vcs"


class _PdmBumpVcsConfig(ConfigNamespace):
    def __init__(
        self, accessor: ConfigAccessor, pdm_bump: "PdmBumpConfig"
    ) -> None:
        super().__init__(_ConfigSections.PDM_BUMP_VCS, accessor, pdm_bump)

    @property
    def provider(self) -> str:
        """"""
        return self._get_value(_ConfigKeys.VCS_PROVIDER)


class PdmBumpConfig(ConfigNamespace):
    """"""

    def __init__(self, accessor: ConfigAccessor) -> None:
        super().__init__(_ConfigSections.PDM_BUMP, accessor, None)
        self.__vcs = _PdmBumpVcsConfig(accessor, self)

    @property
    def vcs(self) -> _PdmBumpVcsConfig:
        """"""
        return self.__vcs

    @property
    def commit_msg_tpl(self) -> str:
        """"""
        return self._get_value("commit_msg_tpl")

    @property
    def perform_commit(self) -> bool:
        """"""
        return self._get_value("perform_commit") or False

    @property
    def auto_tag(self) -> bool:
        """"""
        return self._get_value("auto_tag") or False

    @property
    def tag_allow_dirty(self) -> bool:
        """"""
        return self._get_value("allow_dirty") or False

    @property
    def tag_add_prefix(self) -> bool:
        """"""
        return self._get_value("tag_add_prefix") or True

    def add_values_missing_in_cli(
        self, args: argparse.Namespace
    ) -> argparse.Namespace:
        """"""

        args = PdmBumpConfig.__update_args(args, "commit", self.perform_commit)
        args = PdmBumpConfig.__update_args(
            args, "commit_message", self.commit_msg_tpl
        )
        args = PdmBumpConfig.__update_args(args, "tag", self.auto_tag)
        args = PdmBumpConfig.__update_args(args, "dirty", self.tag_allow_dirty)
        args = PdmBumpConfig.__update_args(
            args, "prepend_letter_v", self.tag_add_prefix
        )

        return args

    @staticmethod
    def __update_args(
        ns: argparse.Namespace, key: str, value: Any
    ) -> argparse.Namespace:
        if key not in ns:
            setattr(ns, key, value)
        else:
            stored_value: Optional[Any] = getattr(ns, key)
            if stored_value is None and value is not None:
                setattr(ns, key, value)
            elif isinstance(stored_value, IsMissing):
                if value is None:
                    setattr(ns, key, stored_value.raw_value())
                else:
                    setattr(ns, key, value)

        return ns


class _PdmBackendConfig:
    """"""

    def __init__(self, accessor: ConfigAccessor) -> None:
        """"""
        self.__mapping = accessor.get_pyproject_config(
            ConfigSection.TOOL_CONFIG
        )

    @property
    def use_scm(self) -> bool:
        """"""
        return (
            self.__mapping.get_config_value(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_SCM
        )

    @property
    def use_file(self) -> bool:
        """"""
        return (
            self.__mapping.get_config_value(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_FILE
        )


class _BuildSystemConfig:
    """"""

    def __init__(self, accessor: ConfigAccessor) -> None:
        """"""
        self.__accessor = accessor
        self.__mapping = accessor.get_pyproject_config(
            ConfigSection.BUILD_SYSTEM
        )
        self.__tool_mapping = accessor.get_pyproject_config(
            ConfigSection.TOOL_CONFIG
        )
        self.__backend = _PdmBackendConfig(accessor)

    @property
    def build_backend(self) -> str:
        """"""
        return self.__mapping.get_config_value(_ConfigKeys.BUILD_BACKEND)

    @property
    def uses_deprecated_build_backed_pdm_pep517(self) -> bool:
        """"""
        return (
            self.build_backend
            == _ConfigValues.DEPRECATED_BUILD_BACKEND_PDM_PEP517_API
            and self.__accessor.get_pyproject_tool_config(
                _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE
            )
            == _ConfigValues.VERSION_SOURCE_FILE
        )

    @property
    def uses_pdm_backend(self) -> bool:
        """"""
        return self.build_backend == _ConfigValues.BUILD_BACKEND_PDM_BACKEND

    @property
    def pdm_backend(self) -> _PdmBackendConfig:
        """"""
        return self.__backend

    @property
    def version_source_file(self) -> Optional[str]:
        """"""
        return self.__tool_mapping.get_config_value(
            _ConfigKeys.VERSION, _ConfigKeys.VERSION_SOURCE_FILE_PATH
        )


class _MetaDataConfig:
    """"""

    def __init__(self, accessor: ConfigAccessor) -> None:
        """"""
        self.__accessor = accessor
        self.__build_system = _BuildSystemConfig(accessor)

    @property
    def version(self) -> Optional[str]:
        """"""
        return self.__accessor.get_pyproject_metadata(_ConfigKeys.VERSION)

    @version.setter
    def version(self, value: str) -> None:
        """"""
        self.__accessor.set_pyproject_metadata(value, _ConfigKeys.VERSION)

    @property
    def build_system(self) -> _BuildSystemConfig:
        """"""
        return self.__build_system

    @property
    def is_dynamic_version(self) -> bool:
        """"""
        return _ConfigKeys.VERSION in self.__accessor.meta_data.dynamic


class _PdmBumpConfigAccessor(ConfigAccessor):
    @property
    def plugin_config_name(self) -> Iterable[str]:
        return {"bump-plugin"}


class _EmptyConfig:
    @property
    def root(self) -> Path:
        """"""
        raise NotImplementedError()

    @root.setter
    def root(self, value: Path) -> None:
        """"""
        raise NotImplementedError()

    # Justification: Fulfill protocol
    @property
    # pylint: disable=C0103
    def PYPROJECT_FILENAME(self) -> str:  # noqa: N802 NOSONAR
        """"""
        raise NotImplementedError()

    # Justification: Fulfill protocol
    @PYPROJECT_FILENAME.setter
    # pylint: disable=C0103
    def PYPROJECT_FILENAME(self, value: str) -> None:  # noqa: N802 NOSONAR
        """"""
        raise NotImplementedError()

    @property
    def config(self) -> Mapping[str, Any]:
        """"""
        return {}


# Justification: Just a protocol implementation
class _DummyConfig:  # pylint: disable=R0903
    @property
    def pyproject_file(self) -> Path:
        """"""
        raise NotImplementedError()


class Config:
    """"""

    def __init__(self, project: ConfigHolder) -> None:
        self.__project: ConfigHolder = project
        accessor: ConfigAccessor = _PdmBumpConfigAccessor(self, project)
        self.__pdm_bump = PdmBumpConfig(accessor)
        self.__meta_data = _MetaDataConfig(accessor)

    @staticmethod
    def get_plugin_config_key_prefix() -> str:
        """"""
        config: ConfigHolder = _EmptyConfig()
        accessor: ConfigAccessor = _PdmBumpConfigAccessor(
            _DummyConfig(), config
        )
        section: ConfigSection = ConfigSection.PLUGIN_CONFIG
        paths: Iterable[str]
        paths = accessor.get_config_section_path(section)

        return ".".join(paths)

    @property
    def pdm_bump(self) -> PdmBumpConfig:
        """"""
        return self.__pdm_bump

    @property
    def meta_data(self) -> _MetaDataConfig:
        """"""
        return self.__meta_data

    @cached_property
    @traced_function
    def pyproject_file(self) -> Path:
        """"""
        return self.__project.root / self.__project.PYPROJECT_FILENAME



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from typing import Final, Optional, Protocol

# MyPy cannot resolve this during pull request
from pdm.project.config import ConfigItem as _ConfigItem  # type: ignore

from .actions.hook import CommitChanges, TagChanges
from .core.config import Config
from .plugin import BumpCommand as _Command


class _CoreLike(Protocol):
    """"""

    def register_command(
        self, command: type[_Command], name: Optional[str] = None
    ) -> None:
        """

        Parameters
        ----------
        command: type[_Command] :

        name: Optional[str] :
             (Default value = None)

        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()

    @staticmethod
    def add_config(name: str, config_item: _ConfigItem) -> None:
        """

        Parameters
        ----------
        name: str :

        config_item: _ConfigItem :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


def main(core: _CoreLike) -> None:
    """

    Parameters
    ----------
    core: _CoreLike :


    Returns
    -------

    """
    core.register_command(_Command)

    config_prefix: Final[str] = Config.get_plugin_config_key_prefix()
    env_var_prefix: Final[str] = "PDM_BUMP_"

    core.add_config(
        f"{config_prefix}.commit_msg_tpl",
        _ConfigItem(
            "The default commit message. Uses templates 'from' and 'to'.",
            CommitChanges.default_commit_message,
        ),
    )
    core.add_config(
        f"{config_prefix}.perform_commit",
        _ConfigItem(
            "If set to true, commit the bumped changes automatically",
            CommitChanges.perform_commit,
        ),
    )
    core.add_config(
        f"{config_prefix}.auto_tag",
        _ConfigItem(
            "Create a tag after bumping and committing the changes",
            TagChanges.do_tag,
        ),
    )
    core.add_config(
        f"{config_prefix}.tag_add_prefix",
        _ConfigItem(
            "Adds the prefix v to the version tag",
            TagChanges.prepend_to_tag,
        ),
    )
    core.add_config(
        f"{config_prefix}.allow_dirty",
        _ConfigItem(
            "Allows tagging the project, if it is dirty",
            TagChanges.allow_dirty,
        ),
    )

    vcs_config_namespace: Final[str] = f"{config_prefix}.vcs"
    vcs_env_var_prefix: Final[str] = f"{env_var_prefix}VCS_"

    core.add_config(
        f"{vcs_config_namespace}.provider",
        _ConfigItem(
            "Configures the VCS Provider to use.",
            "git-cli",
            env_var=f"{vcs_env_var_prefix}PROVIDER",
        ),
    )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel, Chase Sterling.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from dataclasses import dataclass, field
from difflib import unified_diff
from functools import cached_property
from pathlib import Path
from re import M as MultilinePattern  # noqa: N811
from re import Match, Pattern
from re import compile as compile_re
from typing import Final, Optional, cast

from pdm_pfsc.logging import logger

from .core.config import VERSION_CONFIG_KEY_NAME, Config
from .core.version import Pep440VersionFormatter, Version

DEFAULT_REGEX: Final[Pattern[str]] = compile_re(
    r"^__version__\s*=\s*[\"'](?P<version>.+?)[\"']\s*(?:#.*)?$",
    MultilinePattern,
)


@dataclass
class _DiffData:
    original_lines: list[str] = field()
    new_lines: list[str] = field()

    def merge(self, other: "_DiffData") -> "_DiffData":
        """"""
        return _DiffData(self.original_lines, other.new_lines)


VERSION_IDENTIFIER: Final[str] = VERSION_CONFIG_KEY_NAME


class DynamicVersionConfig:
    """"""

    __file: Path
    __pattern: Pattern[str]
    __file_encoding: str

    def __init__(
        self,
        file_path: Path,
        pattern: Pattern[str],
        encoding: str = "utf-8",
    ) -> None:
        self.__file = file_path
        self.__pattern = pattern
        self.__file_encoding = encoding

    @property
    def file(self) -> Path:
        """"""
        return self.__file

    @property
    def pattern(self) -> Pattern[str]:
        """"""
        return self.__pattern

    @cached_property
    def dynamic_version(self) -> Optional[str]:
        """"""
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            match = self.__pattern.search(file_ptr.read())
        if match is not None:
            return match.group(VERSION_IDENTIFIER)
        return None

    def replace_dynamic_version(self, new_version: str) -> _DiffData:
        """

        Parameters
        ----------
        new_version: str :


        Returns
        -------

        """
        with self.__file.open("r", encoding=self.__file_encoding) as file_ptr:
            version_file = file_ptr.read()
            match = self.__pattern.search(version_file)
            if match is None:
                raise ValueError("Failed to fetch version")
            match = cast(Match[str], match)
            version_start, version_end = match.span(VERSION_IDENTIFIER)
            new_version_file = (
                version_file[:version_start]
                + new_version
                + version_file[version_end:]
            )
        with self.__file.open("w", encoding=self.__file_encoding) as file_ptr:
            file_ptr.write(new_version_file)

        return _DiffData(
            version_file.splitlines(),
            new_version_file.splitlines(),
        )

    @staticmethod
    def find_dynamic_config(
        root_path: Path, project_config: Config
    ) -> Optional["DynamicVersionConfig"]:
        """

        Parameters
        ----------
        root_path: Path :

        project_config: Config :


        Returns
        -------

        """
        build_system = project_config.meta_data.build_system
        if build_system.uses_deprecated_build_backed_pdm_pep517:
            logger.warning(
                "Build backend pdm-pep517 is deprecated. Consider upgrading."
            )
            file_path = build_system.version_source_file
            if file_path is not None:
                return DynamicVersionConfig(
                    file_path=root_path / file_path,
                    pattern=DEFAULT_REGEX,
                )
        if project_config.meta_data.build_system.uses_pdm_backend:
            if project_config.meta_data.build_system.pdm_backend.use_scm:
                logger.error(
                    "PDM bump cannot be used if version is fetched from scm"
                )
            if project_config.meta_data.build_system.pdm_backend.use_file:
                file_path = build_system.version_source_file
                if file_path is not None:
                    return DynamicVersionConfig(
                        file_path=root_path / file_path,
                        pattern=DEFAULT_REGEX,
                    )
        return None


# Justification: Implementation of minimal protocol
class DynamicVersionSource:  # pylint: disable=R0903
    """"""

    def __init__(self, project_root: Path, config: Config) -> None:
        self.__project_root = project_root
        self.__config = config
        self.__current_version: Optional[Version] = None
        self.__hunk: Optional[_DiffData] = None

    @property
    def is_enabled(self) -> bool:
        """"""
        return self.__config.meta_data.is_dynamic_version

    @property
    def source_file(self) -> Path:
        """"""
        return self.__get_dynamic_version().file

    @property
    def current_version(self) -> Version:
        """"""
        if self.__current_version is not None:
            return cast(Version, self.__current_version)

        dynamic: DynamicVersionConfig = self.__get_dynamic_version()
        version: Optional[str] = dynamic.dynamic_version
        if version is not None:
            self.__current_version = Version.from_string(version)
            return self.__current_version
        raise ValueError(
            f"Failed to find version in {dynamic.file}. "
            f"Make sure it matches {dynamic.pattern}"
        )

    @current_version.setter
    def current_version(self, version: Version) -> None:
        self.__current_version = version
        ver: str = Pep440VersionFormatter().format(version)
        config: DynamicVersionConfig = self.__get_dynamic_version()
        hunk = config.replace_dynamic_version(ver)
        if self.__hunk is None:
            self.__hunk = hunk
        else:
            self.__hunk = cast(_DiffData, self.__hunk).merge(hunk)

    def __get_dynamic_version(self) -> DynamicVersionConfig:
        dynamic_version: Optional[DynamicVersionConfig] = (
            DynamicVersionConfig.find_dynamic_config(
                self.__project_root, self.__config
            )
        )
        if dynamic_version is not None:
            dynamic: DynamicVersionConfig = cast(
                DynamicVersionConfig, dynamic_version
            )
            return dynamic
        raise ValueError(
            f"Failed to extract dynamic version from {self.__project_root}."
            f" Only pdm-pep517 `file` types are supported."
        )

    def get_source_file_change_hunks(self, repository_root: Path) -> list[str]:
        """"""
        dynamic_version_provider: Path = self.source_file
        relative_file_path: Path = dynamic_version_provider.relative_to(
            str(repository_root)
        )

        if self.__hunk is None:
            return []

        hunk: _DiffData = cast(_DiffData, self.__hunk)

        return list(
            unified_diff(
                hunk.original_lines,
                hunk.new_lines,
                fromfile=str(Path("a") / relative_file_path),
                tofile=str(Path("b") / relative_file_path),
                lineterm="",
            )
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""


from abc import abstractmethod
from argparse import Namespace
from collections.abc import Generator
from dataclasses import dataclass, field
from typing import Any, Protocol, cast

from pdm_pfsc.actions import ActionBase, ActionRegistry
from pdm_pfsc.hook import HookGenerator, HookInfo
from pdm_pfsc.logging import logger

from ..core.config import PdmBumpConfig
from ..core.version import Pep440VersionFormatter, Version
from ..vcs.core import HunkSource, VcsProvider
from .hook import CommitChanges, HookExecutor, TagChanges

_formatter = Pep440VersionFormatter()


# Justification: Just a protocol
class VersionPersister(Protocol):  # pylint: disable=R0903
    """"""

    def save_version(self, version: Version) -> None:
        """

        Parameters
        ----------
        version: Version :


        Returns
        -------

        """
        raise NotImplementedError()


class VersionConsumer(ActionBase[Version]):
    """"""

    def __init__(self, version: Version, **kwargs) -> None:
        super().__init__(**kwargs)
        self.__version = version

    @property
    def current_version(self) -> Version:
        """"""
        return self.__version

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return {"version"}.union(ActionBase.get_allowed_arguments())


class VersionModifier(VersionConsumer):
    """"""

    def __init__(
        self, version: Version, persister: VersionPersister, **kwargs
    ) -> None:
        super().__init__(version, **kwargs)
        self.__persister = persister

    @abstractmethod
    def create_new_version(self) -> Version:
        """"""
        raise NotImplementedError()

    @classmethod
    def generate_hook_infos(cls) -> Generator[HookInfo, None, None]:
        """

        Returns:
        --------
        """
        yield HookInfo(CommitChanges)
        yield HookInfo(TagChanges)

    def _report_new_version(self, next_version: Version) -> None:
        """

        Parameters
        ----------
        next_version: Version :


        Returns
        -------

        """
        logger.info(
            "Performing increment of version: %s -> %s",
            _formatter.format(self.current_version),
            _formatter.format(next_version),
        )

    def run(self, dry_run: bool = False) -> Version:
        """

        Parameters
        ----------
        dry_run: bool :
             (Default value = False)

        Returns
        -------

        """
        next_version: Version = self.create_new_version()
        if next_version is self.current_version:
            return next_version

        if not dry_run:
            self.__persister.save_version(next_version)
        else:
            logger.info("Would write new version %s", next_version)

        return next_version

    @classmethod
    def get_allowed_arguments(cls) -> set[str]:
        """"""
        return {"persister"}.union(VersionConsumer.get_allowed_arguments())

    def __str__(self) -> str:
        return self.__class__.__name__.replace(VersionModifier.__name__, "")


@dataclass(frozen=True)
class ExecutionContext:
    """"""

    version: Version = field()
    persister: VersionPersister = field()
    vcs_provider: VcsProvider = field()
    hunk_source: HunkSource = field()
    config: PdmBumpConfig = field()


class _VersionActions(ActionRegistry[ExecutionContext]):
    def execute(  # pylint: disable=R0913
        self,
        /,
        args: Namespace,
        context: ExecutionContext,
    ) -> None:
        """

        Parameters
        ----------
        / :

        args: Namespace :

        context: ExecutionContext :


        Returns
        -------

        """
        kwargs: dict = {}

        kwargs.update(vars(args))

        kwargs = {
            key: value
            for key, value in kwargs.items()
            if not key.startswith("_")
        }

        dry_run: bool = kwargs.pop("dry_run", False)

        selected_command: str = kwargs.pop("selected_command")

        if selected_command not in self._items:
            raise ValueError(
                f"Failed to get command {selected_command}. "
                + f"Valid values are {', '.join(self._items.keys())}."
            )

        clazz: type[ActionBase] = self._items[selected_command]

        allowed_args: set[str] = clazz.get_allowed_arguments()
        allowed_kwargs: dict[str, Any] = {
            "version": context.version,
            "persister": context.persister,
            "vcs_provider": context.vcs_provider,
        }

        allowed_kwargs = {
            key: value
            for key, value in allowed_kwargs.items()
            if key in allowed_args
        }

        kwargs.update(allowed_kwargs)

        command: "ActionBase" = clazz.create_from_command(**kwargs)
        args = context.config.add_values_missing_in_cli(args)

        executor: HookExecutor = HookExecutor(
            context.hunk_source, context.vcs_provider
        )
        if issubclass(clazz, HookGenerator):
            hook_generator: type[HookGenerator] = cast(
                type[HookGenerator], clazz
            )
            for hook_info in hook_generator.generate_hook_infos():
                executor.register(hook_info.create_hook())

        executor.run((command, context.version), args, dry_run=dry_run)


actions = _VersionActions()
action = actions.register()



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, Protocol

from pdm_pfsc.config import MissingValue
from pdm_pfsc.hook import HookBase, HookExecutorBase
from pdm_pfsc.logging import logger, traced_function

from ..core.version import Pep440VersionFormatter, Version
from ..vcs import HunkSource, VcsProvider


class _Executable(Protocol):  # pylint: disable=R0903
    # Justification: Just a protocol
    """"""

    def run(self, dry_run: bool = False) -> Version:
        """

        Parameters
        ----------
            dry_run :


        Returns
        -------

        """
        raise NotImplementedError()


@dataclass(frozen=True)
class PreHookContext:
    """"""

    vcs_provider: VcsProvider = field()
    version: Version = field()

    @cached_property
    def formatted_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.version)


@dataclass(frozen=True)
class PostHookContext:
    """"""

    vcs_provider: VcsProvider = field()
    hunk_source: HunkSource = field()
    version: Version = field()
    previous_version: Version = field()
    version_changed: bool = field()

    @cached_property
    def formatted_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.version)

    @cached_property
    def formatted_previous_version(self) -> str:
        """

        Returns:
        --------
        """
        return Pep440VersionFormatter().format(self.previous_version)


class HookExecutor(HookExecutorBase[tuple[_Executable, Version]]):
    """"""

    def __init__(
        self, hunk_source: HunkSource, vcs_provider: VcsProvider
    ) -> None:
        """

        Parameters:
        -----------
            hunk_source: HunkSource :
            vcs_provider: VcsProvider :

        Returns:
        --------

        """
        self.__hunk_source = hunk_source
        self.__vcs_provider = vcs_provider
        super().__init__()

    @traced_function
    def run(
        self,
        context: tuple[_Executable, Version],
        args: Namespace,
        dry_run: bool = False,
    ) -> None:
        """

        Parameters:
        -----------
            context: tuple[_Executable, Version] :
            args: Namespace :
            dry_run: bool :

        Returns:
        --------

        """
        (executor, version) = context

        pre_call_ctx: PreHookContext = PreHookContext(
            self.__vcs_provider, version
        )
        for hook in self._hooks:
            hook.pre_action_hook(pre_call_ctx, args)

        old_version = version
        version = executor.run(dry_run)

        post_call_ctx: PostHookContext = PostHookContext(
            self.__vcs_provider,
            self.__hunk_source,
            version,
            old_version,
            old_version != version,
        )
        for hook in self._hooks:
            hook.post_action_hook(post_call_ctx, args)


class CommitChanges(HookBase):
    """"""

    default_commit_message: ClassVar[str] = (
        "chore: Bump version {from} to {to}\n\n"
        "Created a commit with a new version {to}.\n"
        "Previous version was {from}."
    )
    perform_commit: ClassVar[bool] = False

    @traced_function
    def post_action_hook(
        self, context: PostHookContext, args: Namespace
    ) -> None:
        """

        Parameters:
        -----------
            context: PostHookContext :
            args: Namespace :

        Returns:
        --------

        """
        if not context.version_changed:
            logger.debug("Cannot commit changes. Nothing to commit")
            return
        kwargs = vars(args)
        if kwargs.pop("commit", False) and not kwargs.pop("dry_run", False):
            message: str = kwargs.pop(
                "commit_message", CommitChanges.default_commit_message
            )

            f_args: dict[str, str] = {
                "to": context.formatted_version,
                "from": context.formatted_previous_version,
            }

            message = message.format(**f_args)

            context.vcs_provider.check_in_deltas(message, context.hunk_source)

    @traced_function
    def pre_action_hook(
        self, context: PreHookContext, args: Namespace
    ) -> None:
        """

        Parameters:
        -----------
            context: PreHookContext :
            args: Namespace :

        Returns:
        --------

        """
        return  # NOSONAR

    @classmethod
    @traced_function
    def configure(cls, parser: ArgumentParser) -> None:
        """

        Parameters:
        -----------

        parser: ArgumentParser :

        Returns:
        --------

        """
        parser.add_argument(
            "--commit",
            "-c",
            action="store_true",
            default=MissingValue(CommitChanges.perform_commit),
            help="Commit changes to repository. Uses configuration value "
            "'perform_commit' to store default action.",
        )

        parser.add_argument(
            "--message",
            "-m",
            dest="commit_message",
            action="store",
            default=MissingValue(CommitChanges.default_commit_message),
            help="The commit message template. May contain "
            "{from} and {to} as format identifier. Uses configuration value "
            "'commit_msg_tmpl' as configured default value.",
        )


class TagChanges(HookBase):
    """"""

    do_tag: ClassVar[bool] = False
    allow_dirty: ClassVar[bool] = False
    prepend_to_tag: ClassVar[bool] = True

    @traced_function
    def post_action_hook(
        self, context: PostHookContext, args: Namespace
    ) -> None:
        """

        Parameters:
        -----------

            context: PostHookContext :

            args: Namespace :

        Returns:
        --------

        """
        if not context.version_changed:
            return
        kwargs = vars(args)
        if not kwargs.pop("dry_run", False) and kwargs.pop("tag", False):
            if not args.dirty and not context.vcs_provider.is_clean:
                raise RuntimeError(
                    "This should only take place, if "
                    "the git repository is clean"
                )
            context.vcs_provider.create_tag_from_version(
                context.version, kwargs.pop("prepend_letter_v", True)
            )

    @traced_function
    def pre_action_hook(
        self, context: PreHookContext, args: Namespace
    ) -> None:
        """

        Parameters:
        -----------

            context: PreHookContext :

            args: Namespace :

        Returns:
        --------

        """
        kwargs = vars(args)
        if (
            not context.vcs_provider.is_clean
            and not kwargs.pop("dry_run", False)
            and kwargs.pop("tag", False)
        ):
            raise RuntimeError("Repository root is not clean")

    @classmethod
    @traced_function
    def configure(cls, parser: ArgumentParser) -> None:
        """
        Parameters:
        -----------
            parser: ArgumentParser :

        Returns:
        --------

        """
        parser.add_argument(
            "--tag",
            "-t",
            action="store_true",
            default=MissingValue(TagChanges.do_tag),
            help="Create a tag after modifying the current version. Uses "
            "configuration value 'auto_tag' to store its default value.",
        )

        parser.add_argument(
            "--dirty",
            "-d",
            action="store_true",
            default=MissingValue(TagChanges.allow_dirty),
            help="Create a tag, even if the repository is dirty. Uses "
            "configuration value 'allow_dirty' to store its "
            "default value.",
        )

        parser.add_argument(
            "--no-prepend-v",
            dest="prepend_letter_v",
            action="store_false",
            default=MissingValue(TagChanges.prepend_to_tag),
            help="Do not prepend letter v for the tag. Uses "
            "configuration value 'tag_add_prefix' to store its "
            "default value.",
        )



#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021-2023 Carsten Igel.
#
# This file is part of pdm-bump
# (see https://github.com/carstencodes/pdm-bump).
#
# This file is published using the MIT license.
# Refer to LICENSE for more information
#
""""""

from difflib import unified_diff
from pathlib import Path
from typing import Protocol, Union, cast, runtime_checkable

from pdm_pfsc.logging import logger

from .core.config import Config
from .core.version import Pep440VersionFormatter, Version


# Justification: Minimal protocol
class _ProjectWriter(Protocol):  # pylint: disable=R0903
    """"""

    def write(self, show_message: bool) -> None:
        """

        Parameters
        ----------
        show_message: bool :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


@runtime_checkable
class _ProjectWriterClassic(Protocol):  # pylint: disable=R0903
    """"""

    def write_pyproject(self, show_message: bool) -> None:
        """

        Parameters
        ----------
        show_message: bool :


        Returns
        -------

        """
        # Method empty: Only a protocol stub
        raise NotImplementedError()


# Justification: Minimal protocol
@runtime_checkable
class _ProjectWriterHolder(Protocol):  # pylint: disable=R0903
    """"""

    pyproject: _ProjectWriter


# Justification: Implementation of minimal protocol
class StaticPep621VersionSource:  # pylint: disable=R0903
    """"""

    def __init__(
        self,
        _: Union[_ProjectWriterHolder, _ProjectWriterClassic],
        config: Config,
    ) -> None:
        self.__config = config
        self.__original_file_content = self.__load_config_content()

    @property
    def is_enabled(self) -> bool:
        """"""
        return self.__config.meta_data.version is not None

    @property
    def source_file(self) -> Path:
        """"""
        return self.__config.pyproject_file

    @property
    def current_version(self) -> Version:
        """"""
        version: str = cast(str, self.__config.meta_data.version)
        return Version.from_string(version)

    @current_version.setter
    def current_version(self, ver: Version) -> None:
        formatter: Pep440VersionFormatter = Pep440VersionFormatter()
        version: str = formatter.format(ver)
        logger.debug("Setting new version %s", version)
        self.__config.meta_data.version = version

    def get_source_file_change_hunks(self, repository_root: Path) -> list[str]:
        """"""
        relative_path: Path = self.source_file.relative_to(repository_root)
        return list(
            unified_diff(
                self.__original_file_content,
                self.__load_config_content(),
                fromfile=str(Path("a") / relative_path),
                tofile=str(Path("b") / relative_path),
                lineterm="",
            )
        )

    def __load_config_content(self) -> list[str]:
        return self.source_file.read_text().splitlines()
